import mujoco
import numpy as np
from dm_control import mjcf

from ogbench.manipspace.envs.manipspace_env import ManipSpaceEnv


class PuzzleEnv(ManipSpaceEnv):
    """Puzzle environment.

    This environment implements the "Lights Out" puzzle game. The goal is to set all buttons to a specific state. It
    supports the following variants:
    - `env_type`: '3x3', '4x4', '4x5', '4x6'.

    In addition to `qpos` and `qvel`, it maintains the following state variables.
    - `button_states`: A binary array of size `num_buttons` representing the state of each button. Stored in
        `_cur_button_states`.
    """

    def __init__(self, env_type, *args, **kwargs):
        """Initialize the Puzzle environment.

        Args:
            env_type: Environment type. One of '3x3', '4x4', '4x5', or '4x6'.
            *args: Additional arguments to pass to the parent class.
            **kwargs: Additional keyword arguments to pass to the parent class.
        """
        self._env_type = env_type

        # Set the puzzle size.
        self._num_button_states = 2

        if env_type == '3x3':
            self._num_rows = 3
            self._num_cols = 3
        elif env_type == '4x4':
            self._num_rows = 4
            self._num_cols = 4
        elif env_type == '4x5':
            self._num_rows = 4
            self._num_cols = 5
        elif env_type == '4x6':
            self._num_rows = 4
            self._num_cols = 6
        else:
            raise ValueError(f'Unknown env_type: {env_type}')

        self._num_buttons = self._num_rows * self._num_cols
        self._cur_button_states = np.array([0] * self._num_buttons)

        super().__init__(*args, **kwargs)

        # Adjust arm sampling bounds to a smaller region.
        self._arm_sampling_bounds = np.asarray([[0.25, -0.2, 0.20], [0.6, 0.2, 0.25]])

        # Target info.
        self._target_task = 'button'
        self._target_button = 0
        self._target_button_states = np.array([0] * self._num_buttons)

    def set_state(self, qpos, qvel, button_states):
        self._cur_button_states = button_states.copy()
        self._apply_button_states()
        super().set_state(qpos, qvel)

    def set_tasks(self):
        if self._num_rows == 3 and self._num_cols == 3:
            self.task_infos = [
                dict(
                    task_name='task1',
                    init_button_states=np.array(
                        [
                            [0, 0, 0],
                            [0, 0, 0],
                            [0, 0, 0],
                        ]
                    ).flatten(),
                    goal_button_states=np.array(
                        [
                            [1, 1, 0],
                            [1, 0, 1],
                            [0, 1, 1],
                        ]
                    ).flatten(),
                ),
                dict(
                    task_name='task2',
                    init_button_states=np.array(
                        [
                            [1, 1, 1],
                            [1, 1, 1],
                            [1, 1, 1],
                        ]
                    ).flatten(),
                    goal_button_states=np.array(
                        [
                            [0, 1, 1],
                            [1, 1, 1],
                            [1, 1, 1],
                        ]
                    ).flatten(),
                ),
                dict(
                    task_name='task3',
                    init_button_states=np.array(
                        [
                            [0, 1, 0],
                            [1, 1, 1],
                            [0, 1, 0],
                        ]
                    ).flatten(),
                    goal_button_states=np.array(
                        [
                            [1, 0, 1],
                            [0, 1, 0],
                            [1, 0, 1],
                        ]
                    ).flatten(),
                ),
                dict(
                    task_name='task4',
                    init_button_states=np.array(
                        [
                            [0, 1, 0],
                            [1, 0, 1],
                            [0, 1, 0],
                        ]
                    ).flatten(),
                    goal_button_states=np.array(
                        [
                            [1, 1, 1],
                            [1, 1, 1],
                            [1, 1, 1],
                        ]
                    ).flatten(),
                ),
                dict(
                    task_name='task5',
                    init_button_states=np.array(
                        [
                            [1, 1, 1],
                            [1, 1, 1],
                            [1, 1, 1],
                        ]
                    ).flatten(),
                    goal_button_states=np.array(
                        [
                            [1, 0, 1],
                            [1, 0, 1],
                            [1, 0, 1],
                        ]
                    ).flatten(),
                ),
            ]
        elif self._num_rows == 4 and self._num_cols == 4:
            self.task_infos = [
                dict(
                    task_name='task1',
                    init_button_states=np.array(
                        [
                            [0, 0, 0, 0],
                            [0, 0, 0, 0],
                            [0, 0, 0, 0],
                            [0, 0, 0, 0],
                        ]
                    ).flatten(),
                    goal_button_states=np.array(
                        [
                            [1, 1, 1, 1],
                            [1, 1, 1, 1],
                            [1, 1, 1, 1],
                            [1, 1, 1, 1],
                        ]
                    ).flatten(),
                ),
                dict(
                    task_name='task2',
                    init_button_states=np.array(
                        [
                            [1, 1, 1, 1],
                            [1, 1, 1, 1],
                            [1, 1, 1, 1],
                            [1, 1, 1, 1],
                        ]
                    ).flatten(),
                    goal_button_states=np.array(
                        [
                            [1, 1, 1, 1],
                            [1, 0, 0, 1],
                            [1, 0, 0, 1],
                            [1, 1, 1, 1],
                        ]
                    ).flatten(),
                ),
                dict(
                    task_name='task3',
                    init_button_states=np.array(
                        [
                            [1, 0, 0, 0],
                            [0, 0, 0, 0],
                            [0, 0, 0, 0],
                            [0, 0, 0, 1],
                        ]
                    ).flatten(),
                    goal_button_states=np.array(
                        [
                            [0, 1, 0, 1],
                            [1, 0, 1, 0],
                            [0, 1, 0, 1],
                            [1, 0, 1, 0],
                        ]
                    ).flatten(),
                ),
                dict(
                    task_name='task4',
                    init_button_states=np.array(
                        [
                            [1, 0, 0, 1],
                            [1, 0, 0, 1],
                            [1, 0, 0, 1],
                            [1, 0, 0, 1],
                        ]
                    ).flatten(),
                    goal_button_states=np.array(
                        [
                            [0, 0, 0, 0],
                            [0, 0, 0, 0],
                            [0, 0, 0, 0],
                            [0, 0, 0, 0],
                        ]
                    ).flatten(),
                ),
                dict(
                    task_name='task5',
                    init_button_states=np.array(
                        [
                            [0, 1, 0, 1],
                            [0, 0, 1, 0],
                            [0, 0, 0, 1],
                            [1, 0, 0, 0],
                        ]
                    ).flatten(),
                    goal_button_states=np.array(
                        [
                            [0, 0, 0, 0],
                            [0, 0, 0, 0],
                            [0, 0, 0, 0],
                            [0, 0, 0, 0],
                        ]
                    ).flatten(),
                ),
            ]
        elif self._num_rows == 4 and self._num_cols == 5:
            self.task_infos = [
                dict(
                    task_name='task1',
                    init_button_states=np.array(
                        [
                            [1, 1, 0, 1, 1],
                            [0, 1, 0, 1, 0],
                            [0, 1, 0, 1, 0],
                            [1, 1, 0, 1, 1],
                        ]
                    ).flatten(),
                    goal_button_states=np.array(
                        [
                            [0, 0, 0, 0, 0],
                            [0, 0, 0, 0, 0],
                            [0, 0, 0, 0, 0],
                            [0, 0, 0, 0, 0],
                        ]
                    ).flatten(),
                ),
                dict(
                    task_name='task2',
                    init_button_states=np.array(
                        [
                            [0, 0, 0, 0, 0],
                            [0, 0, 0, 0, 0],
                            [0, 0, 0, 0, 0],
                            [0, 0, 0, 0, 0],
                        ]
                    ).flatten(),
                    goal_button_states=np.array(
                        [
                            [1, 1, 1, 1, 1],
                            [1, 1, 1, 1, 1],
                            [1, 1, 1, 1, 1],
                            [1, 1, 1, 1, 1],
                        ]
                    ).flatten(),
                ),
                dict(
                    task_name='task3',
                    init_button_states=np.array(
                        [
                            [0, 0, 0, 0, 0],
                            [0, 0, 1, 0, 0],
                            [0, 0, 1, 0, 0],
                            [0, 0, 0, 0, 0],
                        ]
                    ).flatten(),
                    goal_button_states=np.array(
                        [
                            [1, 1, 1, 1, 1],
                            [1, 0, 0, 0, 1],
                            [1, 0, 0, 0, 1],
                            [1, 1, 1, 1, 1],
                        ]
                    ).flatten(),
                ),
                dict(
                    task_name='task4',
                    init_button_states=np.array(
                        [
                            [0, 0, 0, 0, 0],
                            [0, 0, 0, 0, 0],
                            [0, 0, 0, 0, 0],
                            [0, 0, 0, 0, 0],
                        ]
                    ).flatten(),
                    goal_button_states=np.array(
                        [
                            [0, 0, 0, 0, 0],
                            [0, 0, 1, 0, 0],
                            [0, 0, 1, 0, 0],
                            [0, 0, 0, 0, 0],
                        ]
                    ).flatten(),
                ),
                dict(
                    task_name='task5',
                    init_button_states=np.array(
                        [
                            [0, 0, 0, 0, 0],
                            [0, 0, 0, 0, 0],
                            [0, 0, 0, 0, 0],
                            [0, 0, 0, 0, 0],
                        ]
                    ).flatten(),
                    goal_button_states=np.array(
                        [
                            [1, 0, 0, 0, 1],
                            [0, 1, 1, 1, 0],
                            [0, 1, 1, 1, 0],
                            [1, 0, 0, 0, 1],
                        ]
                    ).flatten(),
                ),
            ]
        elif self._num_rows == 4 and self._num_cols == 6:
            self.task_infos = [
                dict(
                    task_name='task1',
                    init_button_states=np.array(
                        [
                            [1, 1, 0, 1, 1, 1],
                            [0, 0, 1, 0, 1, 0],
                            [0, 1, 0, 1, 0, 0],
                            [1, 1, 1, 0, 1, 1],
                        ]
                    ).flatten(),
                    goal_button_states=np.array(
                        [
                            [0, 0, 0, 0, 0, 0],
                            [0, 0, 0, 0, 0, 0],
                            [0, 0, 0, 0, 0, 0],
                            [0, 0, 0, 0, 0, 0],
                        ]
                    ).flatten(),
                ),
                dict(
                    task_name='task2',
                    init_button_states=np.array(
                        [
                            [1, 1, 1, 1, 1, 1],
                            [1, 1, 1, 1, 1, 1],
                            [1, 1, 1, 1, 1, 1],
                            [1, 1, 1, 1, 1, 1],
                        ]
                    ).flatten(),
                    goal_button_states=np.array(
                        [
                            [0, 0, 0, 0, 0, 0],
                            [0, 0, 0, 0, 0, 0],
                            [0, 0, 0, 0, 0, 0],
                            [0, 0, 0, 0, 0, 0],
                        ]
                    ).flatten(),
                ),
                dict(
                    task_name='task3',
                    init_button_states=np.array(
                        [
                            [0, 0, 0, 0, 0, 0],
                            [0, 0, 0, 0, 0, 0],
                            [0, 0, 0, 0, 0, 0],
                            [0, 0, 0, 0, 0, 0],
                        ]
                    ).flatten(),
                    goal_button_states=np.array(
                        [
                            [1, 1, 1, 1, 1, 0],
                            [1, 1, 0, 1, 0, 1],
                            [1, 0, 1, 0, 1, 1],
                            [0, 1, 1, 1, 1, 1],
                        ]
                    ).flatten(),
                ),
                dict(
                    task_name='task4',
                    init_button_states=np.array(
                        [
                            [0, 1, 0, 1, 0, 1],
                            [1, 0, 1, 0, 1, 0],
                            [0, 1, 0, 1, 0, 1],
                            [1, 0, 1, 0, 1, 0],
                        ]
                    ).flatten(),
                    goal_button_states=np.array(
                        [
                            [1, 0, 0, 0, 0, 1],
                            [0, 0, 0, 0, 0, 0],
                            [0, 0, 0, 0, 0, 0],
                            [1, 0, 0, 0, 0, 1],
                        ]
                    ).flatten(),
                ),
                dict(
                    task_name='task5',
                    init_button_states=np.array(
                        [
                            [0, 0, 0, 0, 0, 0],
                            [0, 0, 0, 0, 0, 0],
                            [0, 0, 0, 0, 0, 0],
                            [0, 0, 0, 0, 0, 0],
                        ]
                    ).flatten(),
                    goal_button_states=np.array(
                        [
                            [1, 0, 0, 0, 0, 1],
                            [0, 1, 1, 1, 1, 0],
                            [0, 1, 1, 1, 1, 0],
                            [1, 0, 0, 0, 0, 1],
                        ]
                    ).flatten(),
                ),
            ]

        if self._reward_task_id == 0:
            # Set default task.
            if self._num_rows == 3 and self._num_cols == 3:
                self._reward_task_id = 4
            elif self._num_rows == 4 and self._num_cols == 4:
                self._reward_task_id = 4
            elif self._num_rows == 4 and self._num_cols == 5:
                self._reward_task_id = 2
            elif self._num_rows == 4 and self._num_cols == 6:
                self._reward_task_id = 2

    def add_objects(self, arena_mjcf):
        # Add button scene.
        button_outer_mjcf = mjcf.from_path((self._desc_dir / 'button_outer.xml').as_posix())
        arena_mjcf.include_copy(button_outer_mjcf)

        # Add buttons to the scene.
        distance = 0.05
        for i in range(self._num_rows):
            for j in range(self._num_cols):
                button_mjcf = mjcf.from_path((self._desc_dir / 'button_inner.xml').as_posix())
                pos_x = 0.425 - distance * (self._num_rows - 1) + 2 * distance * i
                pos_y = 0.0 - distance * (self._num_cols - 1) + 2 * distance * j
                button_mjcf.find('body', 'buttonbox_0').pos[:2] = np.array([pos_x, pos_y])
                for tag in ['body', 'joint', 'geom', 'site']:
                    for item in button_mjcf.find_all(tag):
                        if hasattr(item, 'name') and item.name is not None and item.name.endswith('_0'):
                            item.name = item.name[:-2] + f'_{i * self._num_cols + j}'
                arena_mjcf.include_copy(button_mjcf)

        # Save button geoms.
        self._button_geoms_list = []
        for i in range(self._num_buttons):
            self._button_geoms_list.append([arena_mjcf.find('geom', f'btngeom_{i}')])

        # Add cameras.
        cameras = {
            'front': {
                'pos': (1.139, 0.000, 0.821),
                'xyaxes': (0.000, 1.000, 0.000, -0.627, 0.000, 0.779),
            },
            'front_pixels': {
                'pos': (0.905, 0.000, 0.762),
                'xyaxes': (0.000, 1.000, 0.000, -0.771, 0.000, 0.637),
            },
        }
        for camera_name, camera_kwargs in cameras.items():
            arena_mjcf.worldbody.add('camera', name=camera_name, **camera_kwargs)

    def post_compilation_objects(self):
        # Button geom IDs.
        self._button_geom_ids_list = [
            [self._model.geom(geom.full_identifier).id for geom in button_geoms]
            for button_geoms in self._button_geoms_list
        ]
        self._button_site_ids = [self._model.site(f'btntop_{i}').id for i in range(self._num_buttons)]

    def _apply_button_states(self):
        # Adjust button colors based on the current state.
        for i in range(self._num_buttons):
            for gid in self._button_geom_ids_list[i]:
                color_zero = self._colors['red']
                color_one = self._colors['blue']
                self._model.geom(gid).rgba = color_zero if self._cur_button_states[i] == 0 else color_one

        mujoco.mj_forward(self._model, self._data)

    def initialize_episode(self):
        self._data.qpos[self._arm_joint_ids] = self._home_qpos
        mujoco.mj_kinematics(self._model, self._data)

        if self._mode == 'data_collection':
            # Randomize the scene.

            self.initialize_arm()

            # Randomize button states.
            for i in range(self._num_buttons):
                self._cur_button_states[i] = self.np_random.choice(self._num_button_states)
            self._apply_button_states()

            # Set a new target.
            self.set_new_target(return_info=False)
        else:
            # Set button states based on the current task.

            # Get the current task info.
            init_button_states = self.cur_task_info['init_button_states'].copy()
            goal_button_states = self.cur_task_info['goal_button_states'].copy()

            # First, force set the current scene to the goal state to obtain the goal observation.
            saved_qpos = self._data.qpos.copy()
            saved_qvel = self._data.qvel.copy()
            self.initialize_arm()
            self._cur_button_states = goal_button_states.copy()
            self._apply_button_states()
            mujoco.mj_forward(self._model, self._data)

            # Do a few random steps to make the scene stable.
            for _ in range(5):
                action = self.action_space.sample()
                action[-1] = 1  # Close gripper.
                self.step(action)

            # Save the goal observation.
            self._cur_goal_ob = (
                self.compute_oracle_observation() if self._use_oracle_rep else self.compute_observation()
            )
            if self._render_goal:
                self._cur_goal_rendered = self.render()
            else:
                self._cur_goal_rendered = None

            # Now, do the actual reset.
            self._data.qpos[:] = saved_qpos
            self._data.qvel[:] = saved_qvel
            self.initialize_arm()
            self._cur_button_states = init_button_states.copy()
            self._target_button_states = goal_button_states.copy()
            self._apply_button_states()

        # Forward kinematics to update site positions.
        self.pre_step()
        mujoco.mj_forward(self._model, self._data)
        self.post_step()

        self._success = False

    def set_new_target(self, return_info=True, p_stack=0.5):
        """Set a new random target for data collection.

        Args:
            return_info: Whether to return the observation and reset info.
            p_stack: Unused; defined for compatibility with the other environments.
        """
        assert self._mode == 'data_collection'

        # Set target button.
        self._target_button = self.np_random.choice(self._num_buttons)
        self._target_button_states[self._target_button] = (
            self._cur_button_states[self._target_button] + 1
        ) % self._num_button_states

        mujoco.mj_kinematics(self._model, self._data)

        if return_info:
            return self.compute_observation(), self.get_reset_info()

    def pre_step(self):
        self._prev_button_states = self._cur_button_states.copy()
        super().pre_step()

    def _compute_successes(self):
        """Compute object successes."""
        button_successes = [
            (self._cur_button_states[i] == self._target_button_states[i]) for i in range(self._num_buttons)
        ]

        return button_successes

    def post_step(self):
        # Update button states.
        for i in range(self._num_buttons):
            prev_joint_pos = self._prev_ob_info[f'privileged/button_{i}_pos'][0]
            cur_joint_pos = self._data.joint(f'buttonbox_joint_{i}').qpos.copy()[0]
            if prev_joint_pos > -0.02 and cur_joint_pos <= -0.02:
                # Button pressed: change the state of the button and its neighbors.
                x, y = i // self._num_cols, i % self._num_cols
                for dx, dy in [(0, 0), (1, 0), (-1, 0), (0, 1), (0, -1)]:
                    nx, ny = x + dx, y + dy
                    if 0 <= nx < self._num_rows and 0 <= ny < self._num_cols:
                        self._cur_button_states[nx * self._num_cols + ny] = (
                            self._cur_button_states[nx * self._num_cols + ny] + 1
                        ) % self._num_button_states
        self._apply_button_states()

        # Evaluate successes.
        button_successes = self._compute_successes()
        if self._mode == 'data_collection':
            self._success = button_successes[self._target_button]
        else:
            self._success = all(button_successes)

    def add_object_info(self, ob_info):
        # Button states.
        for i in range(self._num_buttons):
            ob_info[f'privileged/button_{i}_state'] = self._cur_button_states[i]
            ob_info[f'privileged/button_{i}_pos'] = self._data.joint(f'buttonbox_joint_{i}').qpos.copy()
            ob_info[f'privileged/button_{i}_vel'] = self._data.joint(f'buttonbox_joint_{i}').qvel.copy()

        if self._mode == 'data_collection':
            # Target button info.
            ob_info['privileged/target_task'] = self._target_task

            ob_info['privileged/target_button'] = self._target_button
            ob_info['privileged/target_button_state'] = self._target_button_states[self._target_button]
            ob_info['privileged/target_button_top_pos'] = self._data.site_xpos[
                self._button_site_ids[self._target_button]
            ].copy()

        ob_info['prev_button_states'] = self._prev_button_states.copy()
        ob_info['button_states'] = self._cur_button_states.copy()

    def compute_observation(self):
        if self._ob_type == 'pixels':
            return self.get_pixel_observation()
        else:
            xyz_center = np.array([0.425, 0.0, 0.0])
            xyz_scaler = 10.0
            gripper_scaler = 3.0
            button_scaler = 120.0

            ob_info = self.compute_ob_info()
            ob = [
                ob_info['proprio/joint_pos'],
                ob_info['proprio/joint_vel'],
                (ob_info['proprio/effector_pos'] - xyz_center) * xyz_scaler,
                np.cos(ob_info['proprio/effector_yaw']),
                np.sin(ob_info['proprio/effector_yaw']),
                ob_info['proprio/gripper_opening'] * gripper_scaler,
                ob_info['proprio/gripper_contact'],
            ]
            for i in range(self._num_buttons):
                button_state = np.eye(self._num_button_states)[self._cur_button_states[i]]
                ob.extend(
                    [
                        button_state,
                        ob_info[f'privileged/button_{i}_pos'] * button_scaler,
                        ob_info[f'privileged/button_{i}_vel'],
                    ]
                )

            return np.concatenate(ob)

    def compute_oracle_observation(self):
        """Return the oracle goal representation of the current state."""
        return self._cur_button_states.astype(np.float64)

    def compute_reward(self, ob, action):
        if self._reward_task_id is None:
            return super().compute_reward(ob, action)

        # Compute the reward based on the task.
        successes = self._compute_successes()
        reward = float(sum(successes) - len(successes))
        return reward



import contextlib
import os
import warnings

import gymnasium
import mujoco
import numpy as np
from gymnasium import utils
from gymnasium.envs.mujoco import MujocoEnv
from gymnasium.spaces import Box

_DEFAULT_VALUE_AT_MARGIN = 0.1
_STAND_HEIGHT = 1.4
_WALK_SPEED = 1
_RUN_SPEED = 10


def _sigmoids(x, value_at_1, sigmoid):
    if sigmoid in ('cosine', 'linear', 'quadratic'):
        if not 0 <= value_at_1 < 1:
            raise ValueError('`value_at_1` must be nonnegative and smaller than 1, ' 'got {}.'.format(value_at_1))
    else:
        if not 0 < value_at_1 < 1:
            raise ValueError('`value_at_1` must be strictly between 0 and 1, ' 'got {}.'.format(value_at_1))

    if sigmoid == 'gaussian':
        scale = np.sqrt(-2 * np.log(value_at_1))
        return np.exp(-0.5 * (x * scale) ** 2)

    elif sigmoid == 'hyperbolic':
        scale = np.arccosh(1 / value_at_1)
        return 1 / np.cosh(x * scale)

    elif sigmoid == 'long_tail':
        scale = np.sqrt(1 / value_at_1 - 1)
        return 1 / ((x * scale) ** 2 + 1)

    elif sigmoid == 'reciprocal':
        scale = 1 / value_at_1 - 1
        return 1 / (abs(x) * scale + 1)

    elif sigmoid == 'cosine':
        scale = np.arccos(2 * value_at_1 - 1) / np.pi
        scaled_x = x * scale
        with warnings.catch_warnings():
            warnings.filterwarnings(action='ignore', message='invalid value encountered in cos')
            cos_pi_scaled_x = np.cos(np.pi * scaled_x)
        return np.where(abs(scaled_x) < 1, (1 + cos_pi_scaled_x) / 2, 0.0)

    elif sigmoid == 'linear':
        scale = 1 - value_at_1
        scaled_x = x * scale
        return np.where(abs(scaled_x) < 1, 1 - scaled_x, 0.0)

    elif sigmoid == 'quadratic':
        scale = np.sqrt(1 - value_at_1)
        scaled_x = x * scale
        return np.where(abs(scaled_x) < 1, 1 - scaled_x**2, 0.0)

    elif sigmoid == 'tanh_squared':
        scale = np.arctanh(np.sqrt(1 - value_at_1))
        return 1 - np.tanh(x * scale) ** 2

    else:
        raise ValueError('Unknown sigmoid type {!r}.'.format(sigmoid))


def tolerance(x, bounds=(0.0, 0.0), margin=0.0, sigmoid='gaussian', value_at_margin=_DEFAULT_VALUE_AT_MARGIN):
    lower, upper = bounds
    if lower > upper:
        raise ValueError('Lower bound must be <= upper bound.')
    if margin < 0:
        raise ValueError('`margin` must be non-negative.')

    in_bounds = np.logical_and(lower <= x, x <= upper)
    if margin == 0:
        value = np.where(in_bounds, 1.0, 0.0)
    else:
        d = np.where(x < lower, lower - x, x - upper) / margin
        value = np.where(in_bounds, 1.0, _sigmoids(d, value_at_margin, sigmoid))

    return float(value) if np.isscalar(x) else value


class HumanoidEnv(MujocoEnv, utils.EzPickle):
    """DMC Humanoid environment.

    Several methods are reimplemented to remove the dependency on the `dm_control` package. It is supposed to work
    identically to the original Humanoid environment.
    """

    xml_file = os.path.join(os.path.dirname(__file__), 'assets', 'humanoid.xml')
    metadata = {
        'render_modes': ['human', 'rgb_array', 'depth_array'],
        'render_fps': 40,
    }
    if gymnasium.__version__ >= '1.1.0':
        metadata['render_modes'] += ['rgbd_tuple']

    def __init__(
        self,
        xml_file=None,
        task='run',
        **kwargs,
    ):
        if xml_file is None:
            xml_file = self.xml_file
        utils.EzPickle.__init__(
            self,
            xml_file,
            task,
            **kwargs,
        )

        self._move_speed = {
            'walk': _WALK_SPEED,
            'run': _RUN_SPEED,
        }[task]

        observation_space = Box(low=-np.inf, high=np.inf, shape=(67,), dtype=np.float64)

        MujocoEnv.__init__(
            self,
            xml_file,
            frame_skip=5,
            observation_space=observation_space,
            **kwargs,
        )

    def step(self, action):
        self.do_simulation(action, self.frame_skip)

        observation = self._get_obs()
        reward = self._get_reward()

        if self.render_mode == 'human':
            self.render()

        return (
            observation,
            reward,
            False,
            False,
            {
                'xy': self.get_xy(),
            },
        )

    def _step_mujoco_simulation(self, ctrl, n_frames):
        self.data.ctrl[:] = ctrl

        # DMC-style stepping (see Page 6 of https://arxiv.org/abs/2006.12983).
        if self.model.opt.integrator != mujoco.mjtIntegrator.mjINT_RK4.value:
            mujoco.mj_step2(self.model, self.data)
            if n_frames > 1:
                mujoco.mj_step(self.model, self.data, n_frames - 1)
        else:
            mujoco.mj_step(self.model, self.data, n_frames)

        mujoco.mj_step1(self.model, self.data)

    def _get_obs(self):
        joint_angles = self.data.qpos[7:]  # Skip the 7 DoFs of the free root joint.
        head_height = self.data.xpos[2, 2]  # ['head', 'z']
        torso_frame = self.data.xmat[1].reshape(3, 3)  # ['torso']
        torso_pos = self.data.xpos[1]  # ['torso']
        positions = []
        for idx in [16, 10, 13, 7]:  # ['left_hand', 'left_foot', 'right_hand', 'right_foot']
            torso_to_limb = self.data.xpos[idx] - torso_pos
            positions.append(torso_to_limb.dot(torso_frame))
        extremities = np.hstack(positions)
        torso_vertical_orientation = self.data.xmat[1, [6, 7, 8]]  # ['torso', ['zx', 'zy', 'zz']]
        center_of_mass_velocity = self.data.sensordata[0:3]  # ['torso_subtreelinvel']
        velocity = self.data.qvel

        return np.concatenate(
            [joint_angles, [head_height], extremities, torso_vertical_orientation, center_of_mass_velocity, velocity]
        )

    def _get_reward(self):
        head_height = self.data.xpos[2, 2]  # ['head', 'z']
        torso_upright = self.data.xmat[1, 8]  # ['torso', 'zz']
        center_of_mass_velocity = self.data.sensordata[0:3]  # ['torso_subtreelinvel']
        control = self.data.ctrl.copy()

        standing = tolerance(head_height, bounds=(_STAND_HEIGHT, float('inf')), margin=_STAND_HEIGHT / 4)
        upright = tolerance(torso_upright, bounds=(0.9, float('inf')), margin=1.9, sigmoid='linear', value_at_margin=0)
        stand_reward = standing * upright
        small_control = tolerance(control, margin=1, value_at_margin=0, sigmoid='quadratic').mean()
        small_control = (4 + small_control) / 5
        if self._move_speed == 0:
            horizontal_velocity = center_of_mass_velocity[[0, 1]]
            dont_move = tolerance(horizontal_velocity, margin=2).mean()
            return small_control * stand_reward * dont_move
        else:
            com_velocity = np.linalg.norm(center_of_mass_velocity[[0, 1]])
            move = tolerance(
                com_velocity,
                bounds=(self._move_speed, float('inf')),
                margin=self._move_speed,
                value_at_margin=0,
                sigmoid='linear',
            )
            move = (5 * move + 1) / 6
            return small_control * stand_reward * move

    @contextlib.contextmanager
    def disable(self, *flags):
        old_bitmask = self.model.opt.disableflags
        new_bitmask = old_bitmask
        for flag in flags:
            if isinstance(flag, str):
                field_name = 'mjDSBL_' + flag.upper()
                flag = getattr(mujoco.mjtDisableBit, field_name)
            elif isinstance(flag, int):
                flag = mujoco.mjtDisableBit(flag)
            new_bitmask |= flag.value
        self.model.opt.disableflags = new_bitmask
        try:
            yield
        finally:
            self.model.opt.disableflags = old_bitmask

    def reset_model(self):
        penetrating = True
        while penetrating:
            quat = self.np_random.uniform(size=4)
            quat /= np.linalg.norm(quat)
            self.data.qpos[3:7] = quat

            for joint_id in range(1, self.model.njnt):
                range_min, range_max = self.model.jnt_range[joint_id]
                self.data.qpos[6 + joint_id] = self.np_random.uniform(range_min, range_max)

            with self.disable('actuation'):
                mujoco.mj_forward(self.model, self.data)
            penetrating = self.data.ncon > 0

        observation = self._get_obs()

        return observation

    def get_xy(self):
        return self.data.qpos[:2].copy()

    def set_xy(self, xy):
        qpos = self.data.qpos.copy()
        qvel = self.data.qvel.copy()
        qpos[:2] = xy
        self.set_state(qpos, qvel)



from typing import Any, Optional, Sequence

import distrax
import flax
import flax.linen as nn
import jax
import jax.numpy as jnp


def default_init(scale=1.0):
    """Default kernel initializer."""
    return nn.initializers.variance_scaling(scale, 'fan_avg', 'uniform')


def ensemblize(cls, num_qs, out_axes=0, **kwargs):
    """Ensemblize a module."""
    return nn.vmap(
        cls,
        variable_axes={'params': 0},
        split_rngs={'params': True},
        in_axes=None,
        out_axes=out_axes,
        axis_size=num_qs,
        **kwargs,
    )


class Identity(nn.Module):
    """Identity layer."""

    def __call__(self, x):
        return x


class MLP(nn.Module):
    """Multi-layer perceptron.

    Attributes:
        hidden_dims: Hidden layer dimensions.
        activations: Activation function.
        activate_final: Whether to apply activation to the final layer.
        kernel_init: Kernel initializer.
        layer_norm: Whether to apply layer normalization.
    """

    hidden_dims: Sequence[int]
    activations: Any = nn.gelu
    activate_final: bool = False
    kernel_init: Any = default_init()
    layer_norm: bool = False

    @nn.compact
    def __call__(self, x):
        for i, size in enumerate(self.hidden_dims):
            x = nn.Dense(size, kernel_init=self.kernel_init)(x)
            if i + 1 < len(self.hidden_dims) or self.activate_final:
                x = self.activations(x)
                if self.layer_norm:
                    x = nn.LayerNorm()(x)
        return x


class LengthNormalize(nn.Module):
    """Length normalization layer.

    It normalizes the input along the last dimension to have a length of sqrt(dim).
    """

    @nn.compact
    def __call__(self, x):
        return x / jnp.linalg.norm(x, axis=-1, keepdims=True) * jnp.sqrt(x.shape[-1])


class Param(nn.Module):
    """Scalar parameter module."""

    init_value: float = 0.0

    @nn.compact
    def __call__(self):
        return self.param('value', init_fn=lambda key: jnp.full((), self.init_value))


class LogParam(nn.Module):
    """Scalar parameter module with log scale."""

    init_value: float = 1.0

    @nn.compact
    def __call__(self):
        log_value = self.param('log_value', init_fn=lambda key: jnp.full((), jnp.log(self.init_value)))
        return jnp.exp(log_value)


class TransformedWithMode(distrax.Transformed):
    """Transformed distribution with mode calculation."""

    def mode(self):
        return self.bijector.forward(self.distribution.mode())


class RunningMeanStd(flax.struct.PyTreeNode):
    """Running mean and standard deviation.

    Attributes:
        eps: Epsilon value to avoid division by zero.
        mean: Running mean.
        var: Running variance.
        clip_max: Clip value after normalization.
        count: Number of samples.
    """

    eps: Any = 1e-6
    mean: Any = 1.0
    var: Any = 1.0
    clip_max: Any = 10.0
    count: int = 0

    def normalize(self, batch):
        batch = (batch - self.mean) / jnp.sqrt(self.var + self.eps)
        batch = jnp.clip(batch, -self.clip_max, self.clip_max)
        return batch

    def unnormalize(self, batch):
        return batch * jnp.sqrt(self.var + self.eps) + self.mean

    def update(self, batch):
        batch_mean, batch_var = jnp.mean(batch, axis=0), jnp.var(batch, axis=0)
        batch_count = len(batch)

        delta = batch_mean - self.mean
        total_count = self.count + batch_count

        new_mean = self.mean + delta * batch_count / total_count
        m_a = self.var * self.count
        m_b = batch_var * batch_count
        m_2 = m_a + m_b + delta**2 * self.count * batch_count / total_count
        new_var = m_2 / total_count

        return self.replace(mean=new_mean, var=new_var, count=total_count)


class GCActor(nn.Module):
    """Goal-conditioned actor.

    Attributes:
        hidden_dims: Hidden layer dimensions.
        action_dim: Action dimension.
        log_std_min: Minimum value of log standard deviation.
        log_std_max: Maximum value of log standard deviation.
        tanh_squash: Whether to squash the action with tanh.
        state_dependent_std: Whether to use state-dependent standard deviation.
        const_std: Whether to use constant standard deviation.
        final_fc_init_scale: Initial scale of the final fully-connected layer.
        gc_encoder: Optional GCEncoder module to encode the inputs.
    """

    hidden_dims: Sequence[int]
    action_dim: int
    log_std_min: Optional[float] = -5
    log_std_max: Optional[float] = 2
    tanh_squash: bool = False
    state_dependent_std: bool = False
    const_std: bool = True
    final_fc_init_scale: float = 1e-2
    gc_encoder: nn.Module = None

    def setup(self):
        self.actor_net = MLP(self.hidden_dims, activate_final=True)
        self.mean_net = nn.Dense(self.action_dim, kernel_init=default_init(self.final_fc_init_scale))
        if self.state_dependent_std:
            self.log_std_net = nn.Dense(self.action_dim, kernel_init=default_init(self.final_fc_init_scale))
        else:
            if not self.const_std:
                self.log_stds = self.param('log_stds', nn.initializers.zeros, (self.action_dim,))

    def __call__(
        self,
        observations,
        goals=None,
        goal_encoded=False,
        temperature=1.0,
    ):
        """Return the action distribution.

        Args:
            observations: Observations.
            goals: Goals (optional).
            goal_encoded: Whether the goals are already encoded.
            temperature: Scaling factor for the standard deviation.
        """
        if self.gc_encoder is not None:
            inputs = self.gc_encoder(observations, goals, goal_encoded=goal_encoded)
        else:
            inputs = [observations]
            if goals is not None:
                inputs.append(goals)
            inputs = jnp.concatenate(inputs, axis=-1)
        outputs = self.actor_net(inputs)

        means = self.mean_net(outputs)
        if self.state_dependent_std:
            log_stds = self.log_std_net(outputs)
        else:
            if self.const_std:
                log_stds = jnp.zeros_like(means)
            else:
                log_stds = self.log_stds

        log_stds = jnp.clip(log_stds, self.log_std_min, self.log_std_max)

        distribution = distrax.MultivariateNormalDiag(loc=means, scale_diag=jnp.exp(log_stds) * temperature)
        if self.tanh_squash:
            distribution = TransformedWithMode(distribution, distrax.Block(distrax.Tanh(), ndims=1))

        return distribution


class GCDiscreteActor(nn.Module):
    """Goal-conditioned actor for discrete actions.

    Attributes:
        hidden_dims: Hidden layer dimensions.
        action_dim: Action dimension.
        final_fc_init_scale: Initial scale of the final fully-connected layer.
        gc_encoder: Optional GCEncoder module to encode the inputs.
    """

    hidden_dims: Sequence[int]
    action_dim: int
    final_fc_init_scale: float = 1e-2
    gc_encoder: nn.Module = None

    def setup(self):
        self.actor_net = MLP(self.hidden_dims, activate_final=True)
        self.logit_net = nn.Dense(self.action_dim, kernel_init=default_init(self.final_fc_init_scale))

    def __call__(
        self,
        observations,
        goals=None,
        goal_encoded=False,
        temperature=1.0,
    ):
        """Return the action distribution.

        Args:
            observations: Observations.
            goals: Goals (optional).
            goal_encoded: Whether the goals are already encoded.
            temperature: Inverse scaling factor for the logits (set to 0 to get the argmax).
        """
        if self.gc_encoder is not None:
            inputs = self.gc_encoder(observations, goals, goal_encoded=goal_encoded)
        else:
            inputs = [observations]
            if goals is not None:
                inputs.append(goals)
            inputs = jnp.concatenate(inputs, axis=-1)
        outputs = self.actor_net(inputs)

        logits = self.logit_net(outputs)

        distribution = distrax.Categorical(logits=logits / jnp.maximum(1e-6, temperature))

        return distribution


class GCValue(nn.Module):
    """Goal-conditioned value/critic function.

    This module can be used for both value V(s, g) and critic Q(s, a, g) functions.

    Attributes:
        hidden_dims: Hidden layer dimensions.
        layer_norm: Whether to apply layer normalization.
        ensemble: Whether to ensemble the value function.
        gc_encoder: Optional GCEncoder module to encode the inputs.
    """

    hidden_dims: Sequence[int]
    layer_norm: bool = True
    ensemble: bool = True
    gc_encoder: nn.Module = None

    def setup(self):
        mlp_module = MLP
        if self.ensemble:
            mlp_module = ensemblize(mlp_module, 2)
        value_net = mlp_module((*self.hidden_dims, 1), activate_final=False, layer_norm=self.layer_norm)

        self.value_net = value_net

    def __call__(self, observations, goals=None, actions=None):
        """Return the value/critic function.

        Args:
            observations: Observations.
            goals: Goals (optional).
            actions: Actions (optional).
        """
        if self.gc_encoder is not None:
            inputs = [self.gc_encoder(observations, goals)]
        else:
            inputs = [observations]
            if goals is not None:
                inputs.append(goals)
        if actions is not None:
            inputs.append(actions)
        inputs = jnp.concatenate(inputs, axis=-1)

        v = self.value_net(inputs).squeeze(-1)

        return v


class GCDiscreteCritic(GCValue):
    """Goal-conditioned critic for discrete actions."""

    action_dim: int = None

    def __call__(self, observations, goals=None, actions=None):
        actions = jnp.eye(self.action_dim)[actions]
        return super().__call__(observations, goals, actions)


class GCBilinearValue(nn.Module):
    """Goal-conditioned bilinear value/critic function.

    This module computes the value function as V(s, g) = phi(s)^T psi(g) / sqrt(d) or the critic function as
    Q(s, a, g) = phi(s, a)^T psi(g) / sqrt(d), where phi and psi output d-dimensional vectors.

    Attributes:
        hidden_dims: Hidden layer dimensions.
        latent_dim: Latent dimension.
        layer_norm: Whether to apply layer normalization.
        ensemble: Whether to ensemble the value function.
        value_exp: Whether to exponentiate the value. Useful for contrastive learning.
        state_encoder: Optional state encoder.
        goal_encoder: Optional goal encoder.
    """

    hidden_dims: Sequence[int]
    latent_dim: int
    layer_norm: bool = True
    ensemble: bool = True
    value_exp: bool = False
    state_encoder: nn.Module = None
    goal_encoder: nn.Module = None

    def setup(self):
        mlp_module = MLP
        if self.ensemble:
            mlp_module = ensemblize(mlp_module, 2)

        self.phi = mlp_module((*self.hidden_dims, self.latent_dim), activate_final=False, layer_norm=self.layer_norm)
        self.psi = mlp_module((*self.hidden_dims, self.latent_dim), activate_final=False, layer_norm=self.layer_norm)

    def __call__(self, observations, goals, actions=None, info=False):
        """Return the value/critic function.

        Args:
            observations: Observations.
            goals: Goals.
            actions: Actions (optional).
            info: Whether to additionally return the representations phi and psi.
        """
        if self.state_encoder is not None:
            observations = self.state_encoder(observations)
        if self.goal_encoder is not None:
            goals = self.goal_encoder(goals)

        if actions is None:
            phi_inputs = observations
        else:
            phi_inputs = jnp.concatenate([observations, actions], axis=-1)

        phi = self.phi(phi_inputs)
        psi = self.psi(goals)

        v = (phi * psi / jnp.sqrt(self.latent_dim)).sum(axis=-1)

        if self.value_exp:
            v = jnp.exp(v)

        if info:
            return v, phi, psi
        else:
            return v


class GCDiscreteBilinearCritic(GCBilinearValue):
    """Goal-conditioned bilinear critic for discrete actions."""

    action_dim: int = None

    def __call__(self, observations, goals=None, actions=None, info=False):
        actions = jnp.eye(self.action_dim)[actions]
        return super().__call__(observations, goals, actions, info)


class GCMRNValue(nn.Module):
    """Metric residual network (MRN) value function.

    This module computes the value function as the sum of a symmetric Euclidean distance and an asymmetric
    L^infinity-based quasimetric.

    Attributes:
        hidden_dims: Hidden layer dimensions.
        latent_dim: Latent dimension.
        layer_norm: Whether to apply layer normalization.
        encoder: Optional state/goal encoder.
    """

    hidden_dims: Sequence[int]
    latent_dim: int
    layer_norm: bool = True
    encoder: nn.Module = None

    def setup(self):
        self.phi = MLP((*self.hidden_dims, self.latent_dim), activate_final=False, layer_norm=self.layer_norm)

    def __call__(self, observations, goals, is_phi=False, info=False):
        """Return the MRN value function.

        Args:
            observations: Observations.
            goals: Goals.
            is_phi: Whether the inputs are already encoded by phi.
            info: Whether to additionally return the representations phi_s and phi_g.
        """
        if is_phi:
            phi_s = observations
            phi_g = goals
        else:
            if self.encoder is not None:
                observations = self.encoder(observations)
                goals = self.encoder(goals)
            phi_s = self.phi(observations)
            phi_g = self.phi(goals)

        sym_s = phi_s[..., : self.latent_dim // 2]
        sym_g = phi_g[..., : self.latent_dim // 2]
        asym_s = phi_s[..., self.latent_dim // 2 :]
        asym_g = phi_g[..., self.latent_dim // 2 :]
        squared_dist = ((sym_s - sym_g) ** 2).sum(axis=-1)
        quasi = jax.nn.relu((asym_s - asym_g).max(axis=-1))
        v = jnp.sqrt(jnp.maximum(squared_dist, 1e-12)) + quasi

        if info:
            return v, phi_s, phi_g
        else:
            return v


class GCIQEValue(nn.Module):
    """Interval quasimetric embedding (IQE) value function.

    This module computes the value function as an IQE-based quasimetric.

    Attributes:
        hidden_dims: Hidden layer dimensions.
        latent_dim: Latent dimension.
        dim_per_component: Dimension of each component in IQE (i.e., number of intervals in each group).
        layer_norm: Whether to apply layer normalization.
        encoder: Optional state/goal encoder.
    """

    hidden_dims: Sequence[int]
    latent_dim: int
    dim_per_component: int
    layer_norm: bool = True
    encoder: nn.Module = None

    def setup(self):
        self.phi = MLP((*self.hidden_dims, self.latent_dim), activate_final=False, layer_norm=self.layer_norm)
        self.alpha = Param()

    def __call__(self, observations, goals, is_phi=False, info=False):
        """Return the IQE value function.

        Args:
            observations: Observations.
            goals: Goals.
            is_phi: Whether the inputs are already encoded by phi.
            info: Whether to additionally return the representations phi_s and phi_g.
        """
        alpha = jax.nn.sigmoid(self.alpha())
        if is_phi:
            phi_s = observations
            phi_g = goals
        else:
            if self.encoder is not None:
                observations = self.encoder(observations)
                goals = self.encoder(goals)
            phi_s = self.phi(observations)
            phi_g = self.phi(goals)

        x = jnp.reshape(phi_s, (*phi_s.shape[:-1], -1, self.dim_per_component))
        y = jnp.reshape(phi_g, (*phi_g.shape[:-1], -1, self.dim_per_component))
        valid = x < y
        xy = jnp.concatenate(jnp.broadcast_arrays(x, y), axis=-1)
        ixy = xy.argsort(axis=-1)
        sxy = jnp.take_along_axis(xy, ixy, axis=-1)
        neg_inc_copies = jnp.take_along_axis(valid, ixy % self.dim_per_component, axis=-1) * jnp.where(
            ixy < self.dim_per_component, -1, 1
        )
        neg_inp_copies = jnp.cumsum(neg_inc_copies, axis=-1)
        neg_f = -1.0 * (neg_inp_copies < 0)
        neg_incf = jnp.concatenate([neg_f[..., :1], neg_f[..., 1:] - neg_f[..., :-1]], axis=-1)
        components = (sxy * neg_incf).sum(axis=-1)
        v = alpha * components.mean(axis=-1) + (1 - alpha) * components.max(axis=-1)

        if info:
            return v, phi_s, phi_g
        else:
            return v



import functools
from typing import Sequence

import flax.linen as nn
import jax.numpy as jnp

from utils.networks import MLP


class ResnetStack(nn.Module):
    """ResNet stack module."""

    num_features: int
    num_blocks: int
    max_pooling: bool = True

    @nn.compact
    def __call__(self, x):
        initializer = nn.initializers.xavier_uniform()
        conv_out = nn.Conv(
            features=self.num_features,
            kernel_size=(3, 3),
            strides=1,
            kernel_init=initializer,
            padding='SAME',
        )(x)

        if self.max_pooling:
            conv_out = nn.max_pool(
                conv_out,
                window_shape=(3, 3),
                padding='SAME',
                strides=(2, 2),
            )

        for _ in range(self.num_blocks):
            block_input = conv_out
            conv_out = nn.relu(conv_out)
            conv_out = nn.Conv(
                features=self.num_features,
                kernel_size=(3, 3),
                strides=1,
                padding='SAME',
                kernel_init=initializer,
            )(conv_out)

            conv_out = nn.relu(conv_out)
            conv_out = nn.Conv(
                features=self.num_features,
                kernel_size=(3, 3),
                strides=1,
                padding='SAME',
                kernel_init=initializer,
            )(conv_out)
            conv_out += block_input

        return conv_out


class ImpalaEncoder(nn.Module):
    """IMPALA encoder."""

    width: int = 1
    stack_sizes: tuple = (16, 32, 32)
    num_blocks: int = 2
    dropout_rate: float = None
    mlp_hidden_dims: Sequence[int] = (512,)
    layer_norm: bool = False

    def setup(self):
        stack_sizes = self.stack_sizes
        self.stack_blocks = [
            ResnetStack(
                num_features=stack_sizes[i] * self.width,
                num_blocks=self.num_blocks,
            )
            for i in range(len(stack_sizes))
        ]
        if self.dropout_rate is not None:
            self.dropout = nn.Dropout(rate=self.dropout_rate)

    @nn.compact
    def __call__(self, x, train=True, cond_var=None):
        x = x.astype(jnp.float32) / 255.0

        conv_out = x

        for idx in range(len(self.stack_blocks)):
            conv_out = self.stack_blocks[idx](conv_out)
            if self.dropout_rate is not None:
                conv_out = self.dropout(conv_out, deterministic=not train)

        conv_out = nn.relu(conv_out)
        if self.layer_norm:
            conv_out = nn.LayerNorm()(conv_out)
        out = conv_out.reshape((*x.shape[:-3], -1))

        out = MLP(self.mlp_hidden_dims, activate_final=True, layer_norm=self.layer_norm)(out)

        return out


class GCEncoder(nn.Module):
    """Helper module to handle inputs to goal-conditioned networks.

    It takes in observations (s) and goals (g) and returns the concatenation of `state_encoder(s)`, `goal_encoder(g)`,
    and `concat_encoder([s, g])`. It ignores the encoders that are not provided. This way, the module can handle both
    early and late fusion (or their variants) of state and goal information.
    """

    state_encoder: nn.Module = None
    goal_encoder: nn.Module = None
    concat_encoder: nn.Module = None

    @nn.compact
    def __call__(self, observations, goals=None, goal_encoded=False):
        """Returns the representations of observations and goals.

        If `goal_encoded` is True, `goals` is assumed to be already encoded representations. In this case, either
        `goal_encoder` or `concat_encoder` must be None.
        """
        reps = []
        if self.state_encoder is not None:
            reps.append(self.state_encoder(observations))
        if goals is not None:
            if goal_encoded:
                # Can't have both goal_encoder and concat_encoder in this case.
                assert self.goal_encoder is None or self.concat_encoder is None
                reps.append(goals)
            else:
                if self.goal_encoder is not None:
                    reps.append(self.goal_encoder(goals))
                if self.concat_encoder is not None:
                    reps.append(self.concat_encoder(jnp.concatenate([observations, goals], axis=-1)))
        reps = jnp.concatenate(reps, axis=-1)
        return reps


encoder_modules = {
    'impala': ImpalaEncoder,
    'impala_debug': functools.partial(ImpalaEncoder, num_blocks=1, stack_sizes=(4, 4)),
    'impala_small': functools.partial(ImpalaEncoder, num_blocks=1),
    'impala_large': functools.partial(ImpalaEncoder, stack_sizes=(64, 128, 128), mlp_hidden_dims=(1024,)),
}



import abc
import contextlib
from typing import Any, Callable, Optional, SupportsFloat

import gymnasium as gym
import mujoco
import mujoco.viewer
import numpy as np
from dm_control import mjcf

from ogbench.manipspace import mjcf_utils


class CustomMuJoCoEnv(gym.Env, abc.ABC):
    """Custom MuJoCo environment class."""

    def __init__(
        self,
        physics_timestep: float = 0.002,
        control_timestep: float = 0.002,
        render_mode: Optional[str] = None,
        width: int = 200,
        height: int = 200,
    ):
        """Initialize the MuJoCo environment.

        Args:
            physics_timestep: Internal timestep used for physics simulation.
            control_timestep: Timestep used for control. This is the actual control frequency.
            render_mode: Rendering mode. Unused; for compatibility with `gymnasium`.
            width: Width of the rendered image.
            height: Height of the rendered image.
        """
        self._dirty = True
        self._passive_viewer_handle = None
        self._never_compiled = True
        self._reset_next_step = True
        self._mjcf_model: Optional[mjcf.RootElement] = None
        self._model: Optional[mujoco.MjModel] = None
        self._data: Optional[mujoco.MjData] = None
        self._renderer: Optional[mujoco.Renderer] = None
        self._scene_option = mujoco.MjvOption()
        self._camera = mujoco.MjvCamera()
        self._render_height = height
        self._render_width = width

        self.set_timesteps(
            physics_timestep=float(physics_timestep),
            control_timestep=float(control_timestep),
        )

    @abc.abstractmethod
    def build_mjcf_model(self) -> mjcf.RootElement:
        """Build the MJCF model for the environment using the `mjcf` library.

        Returns:
            The root element of the MJCF model.
        """
        raise NotImplementedError

    def modify_mjcf_model(self, mjcf_model: mjcf.RootElement) -> mjcf.RootElement:
        """Modify the MJCF model at the beginning of each episode.

        This is useful for domain randomization or other forms of model modifications that may require recompilation of
        the MjModel and MjData objects. If the operation performed requires recompilation, call `mark_dirty` to force
        recompilation.

        Args:
            mjcf_model: Root element of the MJCF model.

        Returns:
            The root element of the modified MJCF model.
        """
        return mjcf_model

    @abc.abstractmethod
    def initialize_episode(self) -> None:
        """Initialize the environment at the beginning of each episode."""
        raise NotImplementedError

    @abc.abstractmethod
    def compute_observation(self) -> Any:
        """Compute the observation at each timestep.

        Returns:
            A dictionary of observation arrays.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def compute_reward(self, ob, action) -> SupportsFloat:
        """Compute the reward at each timestep."""
        raise NotImplementedError

    def set_control(self, action) -> None:
        """Set the control signal for the actuators at each timestep.

        This simply forwards the action to the underlying actuators. Override this method to provide custom control
        logic such as end-effector Cartesian control.
        """
        self._data.ctrl[:] = action

    def post_compilation(self) -> None:
        """Perform any post-compilation operations.

        This can be useful for caching references to commonly accessed model or data fields. By default, this method
        does nothing.
        """
        pass

    def terminate_episode(self) -> bool:
        """Determine whether the episode should be terminated.

        Can be used to implement custom termination conditions such as task success, and task failure.
        """
        return False

    def truncate_episode(self) -> bool:
        """Determine whether the episode should be truncated.

        Can be used to implement custom truncation conditions such as time limits.
        """
        return False

    def get_reset_info(self) -> dict:
        """Return a dictionary of information to be included in the reset return."""
        return {}

    def get_step_info(self) -> dict:
        """Return a dictionary of information to be included in the step return."""
        return {}

    def pre_step(self) -> None:
        """Perform any pre-step operations.

        This can be useful for saving information. By default, this method does nothing.
        """
        pass

    def post_step(self) -> None:
        """Perform any post-step operations.

        This can be useful for updating the environment state after the simulation has been stepped. By default, this
        method does nothing.
        """
        pass

    def compile_model_and_data(self):
        """Compile the MJCF model into MjModel and MjData objects."""
        getattr(self._mjcf_model.visual, 'global').offwidth = self._render_width
        getattr(self._mjcf_model.visual, 'global').offheight = self._render_height

        self._model = mujoco.MjModel.from_xml_string(
            xml=mjcf_utils.to_string(self._mjcf_model),
            assets=mjcf_utils.get_assets(self._mjcf_model),
        )
        self._data = mujoco.MjData(self._model)

        # Assign the physics timestep.
        self._model.opt.timestep = self._physics_timestep

        mujoco.mj_resetData(self._model, self._data)
        mujoco.mj_forward(self._model, self._data)

        # Make sure the passive viewer is up-to-date.
        if self._passive_viewer_handle is not None:
            self._passive_viewer_handle._sim().load(self._model, self._data, '')

        # Re-initialize the renderer.
        if self._renderer is not None:
            self._renderer.close()
            self._initialize_renderer()

        # Perform any post-compilation operations.
        self.post_compilation()

        # Mark the environment as clean.
        self._dirty = False

    def mark_dirty(self):
        """Mark the environment as dirty, requiring recompilation of the model."""
        self._dirty = True

    def reset(self, seed: int = None, options=None, **kwargs):
        """Reset the environment to the initial state.

        - If this is the first call to `reset`, build the MJCF model with `build_mjcf_model`.
        - Modify the MJCF model by calling `modify_mjcf`.
        - If the environment is dirty, MjModel and MjData objects will be recompiled. Otherwise, compilation will be
            skipped unless this is the first call to `reset`.
        - Reset the simulation with `mujoco.mj_resetData`.
        - Initialize the episode with `initialize_episode`.
        - Compute the first observation with `compute_observation`.
        """
        super().reset(seed=seed, options=options, **kwargs)
        if self._mjcf_model is None:
            self._mjcf_model = self.build_mjcf_model()
        self._mjcf_model = self.modify_mjcf_model(self._mjcf_model)
        if self._dirty or self._never_compiled:
            self.compile_model_and_data()
            self._never_compiled = False
        else:
            mujoco.mj_resetData(self._model, self._data)
            mujoco.mj_forward(self._model, self._data)
        self._reset_next_step = False
        self.initialize_episode()
        mujoco.mj_forward(self._model, self._data)
        ob = self.compute_observation()
        info = self.get_reset_info()
        return ob, info

    def set_state(self, qpos, qvel):
        """Reset the environment to a specific state."""
        assert qpos.shape == (self.model.nq,) and qvel.shape == (self.model.nv,)
        self._data.qpos[:] = np.copy(qpos)
        self._data.qvel[:] = np.copy(qvel)
        if self._model.na == 0:
            self._data.act[:] = None
        mujoco.mj_forward(self._model, self._data)

    def step(self, action):
        """Step the environment forward by one timestep.

        - Set the control signal with `set_control`.
        - Step the simulation with `mujoco.mj_step`.
        - Compute the observation with `compute_observation`.
        - Compute the reward with `compute_reward`.
        - Check if the time limit has been exceeded.
        """
        if self._reset_next_step:
            return self.reset()
        action = np.array(action)
        self.set_control(action)
        self.pre_step()
        mujoco.mj_step(self._model, self._data, nstep=self._n_steps)
        mujoco.mj_rnePostConstraint(self._model, self._data)  # Compute contact forces.
        self.post_step()
        terminated = self.terminate_episode()
        truncated = self.truncate_episode()
        ob = self.compute_observation()
        reward = self.compute_reward(ob, action)
        info = self.get_step_info()
        return ob, reward, terminated, truncated, info

    @property
    def action_space(self):
        """Return the action space for the environment.

        By default, this returns a Box matching the actuators defined in the model. Override this method to provide a
        custom action space.
        """
        if self._model is None:
            self.reset()
        is_limited = self._model.actuator_ctrllimited.ravel().astype(bool)
        ctrlrange = self._model.actuator_ctrlrange
        return gym.spaces.Box(
            low=np.where(is_limited, ctrlrange[:, 0], -mujoco.mjMAXVAL),
            high=np.where(is_limited, ctrlrange[:, 1], mujoco.mjMAXVAL),
            dtype=np.float32,
        )

    @property
    def observation_space(self):
        """Return the observation space for the environment.

        By default, this returns an empty Dict.
        """
        return gym.spaces.Dict({})

    def set_timesteps(self, physics_timestep: float, control_timestep: float) -> None:
        """Set the physics and control timesteps for the environment.

        The physics timestep will be assigned to the MjModel during compilation. The control timestep is used to
        determine the number of physics steps to take per control step.
        """
        # Check timesteps divisible.
        n_steps = control_timestep / physics_timestep
        rounded_n_steps = int(round(n_steps))
        if abs(n_steps - rounded_n_steps) > 1e-6:
            raise ValueError(
                f'Control timestep {control_timestep} should be an integer multiple of '
                f'physics timestep {physics_timestep}.'
            )

        self._physics_timestep = physics_timestep
        self._control_timestep = control_timestep
        self._n_steps = rounded_n_steps

    # Accessors.

    @property
    def model(self) -> mujoco.MjModel:
        """Return the MjModel object."""
        if self._model is None:
            raise ValueError('MjModel object not yet initialized. Call `reset` to initialize.')
        return self._model

    @property
    def data(self) -> mujoco.MjData:
        """Return the MjData object."""
        if self._data is None:
            raise ValueError('MjData object not yet initialized. Call `reset` to initialize.')
        return self._data

    @property
    def mjcf_model(self) -> mjcf.RootElement:
        """Return the root element of the MJCF model."""
        if self._mjcf_model is None:
            raise ValueError('MJCF model not yet initialized. Call `reset` to initialize.')
        return self._mjcf_model

    def physics_timestep(self) -> float:
        """Return the simulation timestep in seconds."""
        return self._physics_timestep

    def control_timestep(self) -> float:
        """Return the control timestep in seconds."""
        return self._control_timestep

    # Visualization.

    def launch_passive_viewer(self, *args, **kwargs):
        """Launch a passive viewer for the environment."""
        if self._passive_viewer_handle is not None:
            raise ValueError('Passive viewer already launched.')
        if self._model is None or self._data is None:
            raise ValueError('Call `reset` before launching the passive viewer.')
        self._passive_viewer_handle = mujoco.viewer.launch_passive(
            self._model,
            self._data,
            show_left_ui=kwargs.pop('show_left_ui', False),
            show_right_ui=kwargs.pop('show_right_ui', False),
            *args,
            **kwargs,
        )

    def sync_passive_viewer(self):
        """Sync the passive viewer with the current state of the environment."""
        if self._passive_viewer_handle is None:
            raise ValueError('Passive viewer not launched.')
        self._passive_viewer_handle.sync()

    def close_passive_viewer(self):
        """Close the passive viewer."""
        if self._passive_viewer_handle is not None:
            self._passive_viewer_handle.close()
            self._passive_viewer_handle = None

    @contextlib.contextmanager
    def passive_viewer(self, *args, **kwargs):
        """Context manager for the passive viewer."""
        self.launch_passive_viewer(*args, **kwargs)
        mujoco.mjv_defaultFreeCamera(self._model, self._passive_viewer_handle.cam)
        with self._passive_viewer_handle.lock():
            self._passive_viewer_handle.opt.flags |= self._scene_option.flags
            self._passive_viewer_handle.opt.geomgroup = self._scene_option.geomgroup
            self._passive_viewer_handle.opt.sitegroup = self._scene_option.sitegroup
            self._passive_viewer_handle.opt.frame = self._scene_option.frame
        yield self._passive_viewer_handle
        self.close_passive_viewer()

    def _initialize_renderer(self):
        """Initialize the renderer."""
        if self._model is None:
            raise ValueError('Call `reset` before rendering.')
        self._renderer = mujoco.Renderer(model=self._model, height=self._render_height, width=self._render_width)
        mujoco.mjv_defaultFreeCamera(self._model, self._camera)

    def render(
        self,
        camera: Any = None,
        depth: bool = False,
        segmentation: bool = False,
        scene_option: Optional[mujoco.MjvOption] = None,
        scene_callback: Optional[Callable[[mujoco.MjvScene], None]] = None,
    ) -> np.ndarray:
        """Render the current state of the environment."""
        if self._model is None or self._data is None:
            raise ValueError('Call `reset` before render.')

        if self._renderer is None:
            self._initialize_renderer()

        if depth and segmentation:
            raise ValueError('Only one of depth or segmentation can be enabled.')
        if depth:
            self._renderer.enable_depth_rendering()
        elif segmentation:
            self._renderer.enable_segmentation_rendering()
        else:
            self._renderer.disable_depth_rendering()
            self._renderer.disable_segmentation_rendering()

        self._renderer.update_scene(
            data=self._data,
            camera=camera or self._camera,
            scene_option=scene_option or self._scene_option,
        )
        if scene_callback is not None:
            scene_callback(self._renderer.scene)

        return self._renderer.render()



from dataclasses import dataclass

from dm_control.viewer import user_input


@dataclass
class KeyCallback:
    reset: bool = False
    pause: bool = False

    def __call__(self, key: int) -> None:
        if key == user_input.KEY_ENTER:
            self.reset = True
        elif key == user_input.KEY_SPACE:
            self.pause = not self.pause



import copy
from typing import Any

import flax
import jax
import jax.numpy as jnp
import ml_collections
import optax
from utils.flax_utils import ModuleDict, TrainState, nonpytree_field
from utils.networks import GCActor, GCValue, LogParam


class SACAgent(flax.struct.PyTreeNode):
    """Soft actor-critic (SAC) agent."""

    rng: Any
    network: Any
    config: Any = nonpytree_field()

    def critic_loss(self, batch, grad_params, rng):
        """Compute the SAC critic loss."""
        next_dist = self.network.select('actor')(batch['next_observations'])
        next_actions, next_log_probs = next_dist.sample_and_log_prob(seed=rng)

        next_qs = self.network.select('target_critic')(batch['next_observations'], actions=next_actions)
        if self.config['min_q']:
            next_q = jnp.min(next_qs, axis=0)
        else:
            next_q = jnp.mean(next_qs, axis=0)

        target_q = batch['rewards'] + self.config['discount'] * batch['masks'] * next_q
        target_q = target_q - self.config['discount'] * batch['masks'] * next_log_probs * self.network.select('alpha')()

        q = self.network.select('critic')(batch['observations'], actions=batch['actions'], params=grad_params)
        critic_loss = jnp.square(q - target_q).mean()

        return critic_loss, {
            'critic_loss': critic_loss,
            'q_mean': q.mean(),
            'q_max': q.max(),
            'q_min': q.min(),
        }

    def actor_loss(self, batch, grad_params, rng):
        """Compute the SAC actor loss."""
        # Actor loss.
        dist = self.network.select('actor')(batch['observations'], params=grad_params)
        actions, log_probs = dist.sample_and_log_prob(seed=rng)

        qs = self.network.select('critic')(batch['observations'], actions=actions)
        if self.config['min_q']:
            q = jnp.min(qs, axis=0)
        else:
            q = jnp.mean(qs, axis=0)

        actor_loss = (log_probs * self.network.select('alpha')() - q).mean()

        # Entropy loss.
        alpha = self.network.select('alpha')(params=grad_params)
        entropy = -jax.lax.stop_gradient(log_probs).mean()
        alpha_loss = (alpha * (entropy - self.config['target_entropy'])).mean()

        total_loss = actor_loss + alpha_loss

        if self.config['tanh_squash']:
            action_std = dist._distribution.stddev()
        else:
            action_std = dist.stddev().mean()

        return total_loss, {
            'total_loss': total_loss,
            'actor_loss': actor_loss,
            'alpha_loss': alpha_loss,
            'alpha': alpha,
            'entropy': -log_probs.mean(),
            'std': action_std.mean(),
        }

    @jax.jit
    def total_loss(self, batch, grad_params, rng=None):
        """Compute the total loss."""
        info = {}
        rng = rng if rng is not None else self.rng

        rng, actor_rng, critic_rng = jax.random.split(rng, 3)

        critic_loss, critic_info = self.critic_loss(batch, grad_params, critic_rng)
        for k, v in critic_info.items():
            info[f'critic/{k}'] = v

        actor_loss, actor_info = self.actor_loss(batch, grad_params, actor_rng)
        for k, v in actor_info.items():
            info[f'actor/{k}'] = v

        loss = critic_loss + actor_loss
        return loss, info

    def target_update(self, network, module_name):
        """Update the target network."""
        new_target_params = jax.tree_util.tree_map(
            lambda p, tp: p * self.config['tau'] + tp * (1 - self.config['tau']),
            self.network.params[f'modules_{module_name}'],
            self.network.params[f'modules_target_{module_name}'],
        )
        network.params[f'modules_target_{module_name}'] = new_target_params

    @jax.jit
    def update(self, batch):
        """Update the agent and return a new agent with information dictionary."""
        new_rng, rng = jax.random.split(self.rng)

        def loss_fn(grad_params):
            return self.total_loss(batch, grad_params, rng=rng)

        new_network, info = self.network.apply_loss_fn(loss_fn=loss_fn)
        self.target_update(new_network, 'critic')

        return self.replace(network=new_network, rng=new_rng), info

    @jax.jit
    def sample_actions(
        self,
        observations,
        goals=None,
        seed=None,
        temperature=1.0,
    ):
        """Sample actions from the actor."""
        dist = self.network.select('actor')(observations, goals, temperature=temperature)
        actions = dist.sample(seed=seed)
        actions = jnp.clip(actions, -1, 1)
        return actions

    @classmethod
    def create(
        cls,
        seed,
        ex_observations,
        ex_actions,
        config,
    ):
        """Create a new agent.

        Args:
            seed: Random seed.
            ex_observations: Example batch of observations.
            ex_actions: Example batch of actions.
            config: Configuration dictionary.
        """
        rng = jax.random.PRNGKey(seed)
        rng, init_rng = jax.random.split(rng, 2)

        action_dim = ex_actions.shape[-1]

        if config['target_entropy'] is None:
            config['target_entropy'] = -config['target_entropy_multiplier'] * action_dim

        # Define critic and actor networks.
        critic_def = GCValue(
            hidden_dims=config['value_hidden_dims'],
            layer_norm=config['layer_norm'],
            ensemble=True,
        )

        actor_def = GCActor(
            hidden_dims=config['actor_hidden_dims'],
            action_dim=action_dim,
            log_std_min=-5,
            tanh_squash=config['tanh_squash'],
            state_dependent_std=config['state_dependent_std'],
            const_std=False,
            final_fc_init_scale=config['actor_fc_scale'],
        )

        # Define the dual alpha variable.
        alpha_def = LogParam()

        network_info = dict(
            critic=(critic_def, (ex_observations, None, ex_actions)),
            target_critic=(copy.deepcopy(critic_def), (ex_observations, None, ex_actions)),
            actor=(actor_def, (ex_observations, None)),
            alpha=(alpha_def, ()),
        )
        networks = {k: v[0] for k, v in network_info.items()}
        network_args = {k: v[1] for k, v in network_info.items()}

        network_def = ModuleDict(networks)
        network_tx = optax.adam(learning_rate=config['lr'])
        network_params = network_def.init(init_rng, **network_args)['params']
        network = TrainState.create(network_def, network_params, tx=network_tx)

        params = network.params
        params['modules_target_critic'] = params['modules_critic']

        return cls(rng, network=network, config=flax.core.FrozenDict(**config))


def get_config():
    config = ml_collections.ConfigDict(
        dict(
            agent_name='sac',  # Agent name.
            lr=1e-4,  # Learning rate.
            batch_size=256,  # Batch size.
            actor_hidden_dims=(256, 256),  # Actor network hidden dimensions.
            value_hidden_dims=(256, 256),  # Value network hidden dimensions.
            layer_norm=False,  # Whether to use layer normalization.
            discount=0.99,  # Discount factor.
            tau=0.005,  # Target network update rate.
            target_entropy=ml_collections.config_dict.placeholder(float),  # Target entropy (None for automatic tuning).
            target_entropy_multiplier=0.5,  # Multiplier to dim(A) for target entropy.
            tanh_squash=True,  # Whether to squash actions with tanh.
            state_dependent_std=True,  # Whether to use state-dependent standard deviations for actor.
            actor_fc_scale=0.01,  # Final layer initialization scale for actor.
            min_q=True,  # Whether to use min Q (True) or mean Q (False).
        )
    )
    return config



import numpy as np

from ogbench.manipspace.oracles.markov.markov_oracle import MarkovOracle


class WindowMarkovOracle(MarkovOracle):
    def __init__(self, max_step=75, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._max_step = max_step

    def reset(self, ob, info):
        self._done = False
        self._step = 0
        arm_sampling_bounds = self._env.unwrapped._arm_sampling_bounds.copy()
        arm_sampling_bounds[0, 2] = max(arm_sampling_bounds[0, 2], 0.3)
        self._final_pos = np.random.uniform(*arm_sampling_bounds)
        self._final_yaw = np.random.uniform(-np.pi, np.pi)

    def select_action(self, ob, info):
        effector_pos = info['proprio/effector_pos']
        effector_yaw = info['proprio/effector_yaw'][0]
        gripper_opening = info['proprio/gripper_opening']

        window_pos = info['privileged/window_handle_pos']
        window_yaw = self.shortest_yaw(effector_yaw, info['privileged/window_handle_yaw'][0], n=2)
        target_pos = info['privileged/target_window_handle_pos']

        window_above_offset = np.array([0, 0, 0.06])
        window_handle_offset = np.array([0, 0, 0])
        above_threshold = 0.28
        above = effector_pos[2] > above_threshold
        xy_aligned = np.linalg.norm(window_pos[:2] + window_handle_offset[:2] - effector_pos[:2]) <= 0.04
        pos_aligned = np.linalg.norm(window_pos + window_handle_offset - effector_pos) <= 0.03
        target_pos_aligned = np.linalg.norm(target_pos - window_pos) <= 0.01
        final_pos_aligned = np.linalg.norm(self._final_pos - effector_pos) <= 0.04

        gain_pos = 5
        gain_yaw = 3
        action = np.zeros(5)
        if not target_pos_aligned:
            if not xy_aligned:
                self.print_phase('1: Move above the window handle')
                action = np.zeros(5)
                diff = window_pos + window_handle_offset + window_above_offset - effector_pos
                diff = self.shape_diff(diff)
                action[:3] = diff[:3] * gain_pos
                action[3] = (window_yaw - effector_yaw) * gain_yaw
                action[4] = -1
            elif not pos_aligned:
                self.print_phase('2: Move to the window handle')
                diff = window_pos + window_handle_offset - effector_pos
                diff = self.shape_diff(diff)
                action[:3] = diff[:3] * gain_pos
                action[3] = (window_yaw - effector_yaw) * gain_yaw
                action[4] = -1
            else:
                self.print_phase('3: Move to the target')
                diff = target_pos + window_handle_offset - effector_pos
                diff = self.shape_diff(diff)
                action[:3] = diff[:3] * gain_pos
                action[3] = (window_yaw - effector_yaw) * gain_yaw
                action[4] = 1
        else:
            if not above:
                self.print_phase('4: Move in the air')
                diff = (
                    np.array(
                        [
                            window_pos[0],
                            window_pos[1],
                            above_threshold * 2,
                        ]
                    )
                    - effector_pos
                )
                diff = self.shape_diff(diff)
                action[:3] = diff[:3] * gain_pos
                action[3] = (self._final_yaw - effector_yaw) * gain_yaw
                action[4] = -1
            else:
                self.print_phase('5: Move to the final position')
                diff = self._final_pos - effector_pos
                diff = self.shape_diff(diff)
                action[:3] = diff[:3] * gain_pos
                action[3] = (self._final_yaw - effector_yaw) * gain_yaw
                action[4] = -1

            if final_pos_aligned:
                self._done = True

        action = np.clip(action, -1, 1)
        if self._debug:
            print(action)

        self._step += 1
        if self._step == self._max_step:
            self._done = True

        return action



import numpy as np


class Behavior:
    """Base class for action behaviors."""

    def __init__(self, env):
        self._env = env
        self._done = False
        self._step = 0

        self._size = self._env.unwrapped._world_size // self._env.unwrapped._brush_size
        assert self._env.unwrapped._brush_size == self._env.unwrapped._grid_size
        self._elem_name = None
        self._sequence = None

    @property
    def done(self):
        return self._done

    def reset(self, ob, info):
        pass

    def select_action(self, ob, info):
        x, y = self._sequence[self._step]

        self._step += 1
        if self._step == len(self._sequence):
            self._done = True

        return self._elem_name, x, y


class FillBehavior(Behavior):
    """Fill the entire grid with a single element."""

    def reset(self, ob, info):
        self._done = False
        self._step = 0
        self._elem_name = np.random.choice(self._env.unwrapped._elem_names)

        # Randomly flip the fill directions.
        flip_x = np.random.randint(2)
        flip_y = np.random.randint(2)
        flip_xy = np.random.randint(2)

        self._sequence = []
        for i in range(self._size * self._size):
            x = i % self._size
            if flip_x:
                x = self._size - x - 1
            y = i // self._size
            if flip_y:
                y = self._size - y - 1
            if flip_xy:
                x, y = y, x

            self._sequence.append((x, y))


class LineBehavior(Behavior):
    """Fill a single line with a single element."""

    def reset(self, ob, info):
        self._done = False
        self._step = 0
        self._elem_name = np.random.choice(self._env.unwrapped._elem_names)

        # Randomly select the line direction.
        target_idx = np.random.randint(self._size)
        flip_dir = np.random.randint(2)
        flip_xy = np.random.randint(2)

        self._sequence = []
        for i in range(self._size):
            x, y = i, target_idx
            if flip_dir:
                y = self._size - 1 - y
            if flip_xy:
                x, y = y, x

            self._sequence.append((x, y))


class SquareBehavior(Behavior):
    """Draw a square with a single element."""

    def reset(self, ob, info):
        self._done = False
        self._step = 0
        self._elem_name = np.random.choice(self._env.unwrapped._elem_names)

        length = np.random.randint(1, self._size)
        x1 = np.random.randint(self._size - length)
        x2 = x1 + length
        y1 = np.random.randint(self._size - length)
        y2 = y1 + length

        sides = []
        sides.append([(x1, y) for y in range(y1, y2 + 1)])
        sides.append([(x2, y) for y in range(y1, y2 + 1)])
        sides.append([(x, y1) for x in range(x1, x2 + 1)])
        sides.append([(x, y2) for x in range(x1, x2 + 1)])

        # Randomly reverse sides.
        for i in range(4):
            if np.random.randint(2):
                sides[i].reverse()

        # Randomly shuffle the order of sides.
        np.random.shuffle(sides)

        self._sequence = []
        for side in sides:
            self._sequence.extend(side)



import numpy as np

from ogbench.manipspace.oracles.plan.plan_oracle import PlanOracle


class ButtonPlanOracle(PlanOracle):
    def __init__(self, gripper_always_closed=False, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._gripper_always_closed = gripper_always_closed

    def compute_keyframes(self, plan_input):
        # Poses.
        poses = {}
        poses['initial'] = plan_input['effector_initial']
        poses['press_start'] = self.above(plan_input['button'], 0.06)
        poses['press'] = self.above(plan_input['button'], -0.025)
        poses['press_end'] = poses['press_start']
        poses['final'] = plan_input['effector_goal']

        # Times.
        times = {}
        distance = np.linalg.norm(poses['initial'].translation() - poses['press_start'].translation())
        times['initial'] = 0.0
        times['press_start'] = times['initial'] + self._dt * (0.5 + distance * 4)
        times['press'] = times['press_start'] + self._dt * 0.8
        times['press_end'] = times['press'] + self._dt * 0.8
        times['final'] = times['press_end'] + self._dt * 1.25
        for time in times.keys():
            if time != 'initial':
                times[time] += np.random.uniform(-1, 1) * self._dt * 0.1

        # Grasps.
        grasps = {}
        if self._gripper_always_closed:
            g = 1.0
        else:
            g = 0.0
        for name in times.keys():
            if not self._gripper_always_closed:
                if name in {'press_start', 'final'}:
                    g = 1.0 - g
            grasps[name] = g

        return times, poses, grasps

    def reset(self, ob, info):
        plan_input = {
            'effector_initial': self.to_pose(
                pos=info['proprio/effector_pos'],
                yaw=info['proprio/effector_yaw'][0],
            ),
            'effector_goal': self.to_pose(
                pos=np.random.uniform(*self._env.unwrapped._arm_sampling_bounds),
                yaw=np.random.uniform(-np.pi, np.pi),
            ),
            'button': self.to_pose(
                pos=info['privileged/target_button_top_pos'],
                yaw=info['privileged/target_button_top_pos'][0],
            ),
        }

        times, poses, grasps = self.compute_keyframes(plan_input)
        poses = [poses[name] for name in times.keys()]
        grasps = [grasps[name] for name in times.keys()]
        times = list(times.values())

        self._t_init = info['time'][0]
        self._t_max = times[-1]
        self._done = False
        self._plan = self.compute_plan(times, poses, grasps)



import json
import os
import random
import time

import jax
import numpy as np
import tqdm
import wandb
from absl import app, flags
from agents import agents
from ml_collections import config_flags
from online_env_utils import make_online_env
from utils.datasets import ReplayBuffer
from utils.evaluation import evaluate, flatten
from utils.flax_utils import restore_agent, save_agent
from utils.log_utils import CsvLogger, get_exp_name, get_flag_dict, get_wandb_video, setup_wandb
from viz_utils import visualize_trajs

FLAGS = flags.FLAGS

flags.DEFINE_string('run_group', 'Debug', 'Run group.')
flags.DEFINE_integer('seed', 0, 'Random seed.')
flags.DEFINE_string('env_name', 'online-ant-xy-v0', 'Environment name.')
flags.DEFINE_string('save_dir', 'exp/', 'Save directory.')
flags.DEFINE_string('restore_path', None, 'Restore path.')
flags.DEFINE_integer('restore_epoch', None, 'Restore epoch.')

flags.DEFINE_integer('seed_steps', 10000, 'Number of seed steps.')
flags.DEFINE_integer('train_steps', 1000000, 'Number of training steps.')
flags.DEFINE_integer('train_interval', 1, 'Train interval.')
flags.DEFINE_integer('num_epochs', 1, 'Number of updates per train interval.')
flags.DEFINE_integer('log_interval', 5000, 'Logging interval.')
flags.DEFINE_integer('eval_interval', 100000, 'Evaluation interval.')
flags.DEFINE_integer('save_interval', 1000000, 'Saving interval.')
flags.DEFINE_integer('reset_interval', 0, 'Full parameter reset interval.')
flags.DEFINE_integer('terminate_at_end', 0, 'Whether to set terminated=True when truncated=True.')

flags.DEFINE_integer('eval_episodes', 50, 'Number of episodes for each task.')
flags.DEFINE_float('eval_temperature', 0, 'Actor temperature for evaluation.')
flags.DEFINE_float('eval_gaussian', None, 'Action Gaussian noise for evaluation.')
flags.DEFINE_integer('video_episodes', 1, 'Number of video episodes for each task.')
flags.DEFINE_integer('video_frame_skip', 3, 'Frame skip for videos.')
flags.DEFINE_integer('eval_on_cpu', 1, 'Whether to evaluate on CPU.')

config_flags.DEFINE_config_file('agent', '../impls/agents/sac.py', lock_config=False)


def main(_):
    # Set up logger.
    exp_name = get_exp_name(FLAGS.seed)
    setup_wandb(project='OGBench', group=FLAGS.run_group, name=exp_name)

    FLAGS.save_dir = os.path.join(FLAGS.save_dir, wandb.run.project, FLAGS.run_group, exp_name)
    os.makedirs(FLAGS.save_dir, exist_ok=True)
    flag_dict = get_flag_dict()
    with open(os.path.join(FLAGS.save_dir, 'flags.json'), 'w') as f:
        json.dump(flag_dict, f)

    config = FLAGS.agent

    # Set up environments and replay buffer.
    env = make_online_env(FLAGS.env_name)
    eval_env = make_online_env(FLAGS.env_name)

    example_transition = dict(
        observations=env.observation_space.sample(),
        actions=env.action_space.sample(),
        rewards=0.0,
        masks=1.0,
        next_observations=env.observation_space.sample(),
    )

    replay_buffer = ReplayBuffer.create(example_transition, size=int(1e6))

    # Initialize agent.
    random.seed(FLAGS.seed)
    np.random.seed(FLAGS.seed)

    agent_class = agents[config['agent_name']]
    agent = agent_class.create(
        FLAGS.seed,
        example_transition['observations'],
        example_transition['actions'],
        config,
    )

    # Restore agent.
    if FLAGS.restore_path is not None:
        agent = restore_agent(agent, FLAGS.restore_path, FLAGS.restore_epoch)

    # Train agent.
    expl_metrics = dict()
    expl_rng = jax.random.PRNGKey(FLAGS.seed)
    ob, _ = env.reset()

    train_logger = CsvLogger(os.path.join(FLAGS.save_dir, 'train.csv'))
    eval_logger = CsvLogger(os.path.join(FLAGS.save_dir, 'eval.csv'))
    first_time = time.time()
    last_time = time.time()
    update_info = None
    for i in tqdm.tqdm(range(1, FLAGS.train_steps + 1), smoothing=0.1, dynamic_ncols=True):
        # Sample transition.
        if i < FLAGS.seed_steps:
            action = env.action_space.sample()
        else:
            expl_rng, key = jax.random.split(expl_rng)
            action = agent.sample_actions(observations=ob, seed=key)

        action = np.array(action)
        next_ob, reward, terminated, truncated, info = env.step(action)
        if FLAGS.terminate_at_end and truncated:
            terminated = True

        replay_buffer.add_transition(
            dict(
                observations=ob,
                actions=action,
                rewards=reward,
                masks=float(not terminated),
                next_observations=next_ob,
            )
        )
        ob = next_ob

        if terminated or truncated:
            expl_metrics = {f'exploration/{k}': np.mean(v) for k, v in flatten(info).items()}
            ob, _ = env.reset()

        if replay_buffer.size < FLAGS.seed_steps:
            continue

        # Update agent.
        if i % FLAGS.train_interval == 0:
            for _ in range(FLAGS.num_epochs):
                batch = replay_buffer.sample(config['batch_size'])
                agent, update_info = agent.update(batch)

        # Log metrics.
        if i % FLAGS.log_interval == 0 and update_info is not None:
            train_metrics = {f'training/{k}': v for k, v in update_info.items()}
            train_metrics['time/epoch_time'] = (time.time() - last_time) / FLAGS.log_interval
            train_metrics['time/total_time'] = time.time() - first_time
            train_metrics.update(expl_metrics)
            last_time = time.time()
            wandb.log(train_metrics, step=i)
            train_logger.log(train_metrics, step=i)

        # Evaluate agent.
        if i % FLAGS.eval_interval == 0:
            if FLAGS.eval_on_cpu:
                eval_agent = jax.device_put(agent, device=jax.devices('cpu')[0])
            else:
                eval_agent = agent
            eval_metrics = {}
            eval_info, trajs, renders = evaluate(
                agent=eval_agent,
                env=eval_env,
                task_id=None,
                config=config,
                num_eval_episodes=FLAGS.eval_episodes,
                num_video_episodes=FLAGS.video_episodes,
                video_frame_skip=FLAGS.video_frame_skip,
                eval_temperature=FLAGS.eval_temperature,
                eval_gaussian=FLAGS.eval_gaussian,
            )
            eval_metrics.update({f'evaluation/{k}': v for k, v in eval_info.items()})

            if FLAGS.video_episodes > 0:
                video = get_wandb_video(renders=renders)
                eval_metrics['video'] = video

            traj_image = visualize_trajs(FLAGS.env_name, trajs)
            if traj_image is not None:
                eval_metrics['traj'] = wandb.Image(traj_image)

            wandb.log(eval_metrics, step=i)
            eval_logger.log(eval_metrics, step=i)

        # Save agent.
        if i % FLAGS.save_interval == 0:
            save_agent(agent, FLAGS.save_dir, i)

        # Reset agent.
        if FLAGS.reset_interval > 0 and i % FLAGS.reset_interval == 0:
            new_agent = agent_class.create(
                FLAGS.seed + i,
                example_transition['observations'],
                example_transition['actions'],
                config,
            )
            agent = agent.replace(
                network=agent.network.replace(params=new_agent.network.params, opt_state=new_agent.network.opt_state)
            )
            del new_agent
    train_logger.close()
    eval_logger.close()


if __name__ == '__main__':
    app.run(main)
